# $Id: freedb.rb,v 1.6 2002/10/13 21:23:50 moumar Exp $ 

class FreedbError < StandardError ; end

class Freedb
  
  VERSION = "0.4.1"
  PROTO_LEVEL = 5
  CD_FRAME = 75
  
  attr_reader :results, :genre, :exact_genre, :year, :artist, :title, :tracks_ext, :tracks, :length, :discid, :ext_infos, :query

  def initialize(param = "/dev/cdrom", is_query = false)
    @query = 
      if is_query 
	if not param.kind_of? String
	  raise TypeError
	end
        param
      else
        require "freedb_cdrom"
        get_cdrom(param)
      end
    q = @query.split(" ")
    @discid = q[0]
    nb_tracks = q[1].to_i
    @length = q[-1].to_i
    offsets = q[2...-1] << @length*CD_FRAME
    offsets.collect! { |x| x.to_i }

    #@tracks_ext = Array.new(nb_tracks)
    @tracks = Array.new(nb_tracks) # for compatibility
    @tracks_ext = Array.new

    nb_tracks.times { |i|
      t = Hash.new
      t["length"] = (offsets[i+1]-offsets[i])/CD_FRAME
      @tracks_ext << t
    }
  end

  
  def fetch_net(server = "freedb.org", port = 8880)
    @handler = FetchNet.new(server, port)
    _fetch
  end
  alias fetch fetch_net

  def fetch_cgi(server = "www.freedb.org", port = 80, proxy = nil, proxy_port = nil, address = "/~cddb/cddb.cgi")
    @handler = FetchCGI.new(server, port, proxy, proxy_port, address)
    _fetch
  end

  def get_result(index)
    #line = /\w+ \w+/.match(@results[index])[0]
    #puts line
    #@handler.send_cmd("read", line)
  #  @tracks.clear
    @results[index] =~ /\S+ \S+/
    @handler.send_cmd("read", $&)

    # swallow the whole response into a hash
    response = Hash.new

    @handler.each_line { |line|
      case line
	when /^(\d+) (\S+)/, /^([A-Za-z0-9_]+)=(.*)/
	  key = $1.upcase

	  val = $2.gsub(/\\(.)/) {
	    case $1
	      when "t"
		"\t"
	      when "n"
		"\n"
	      else
		$1
	    end
          }

	  (response[key] ||= '') << val
      end
    }

    @genre = response['210']
    @exact_genre = response['DGENRE']
    @year = response['DYEAR'].to_i
    @ext_infos = response['EXTD']

    # Use a regexp instead of a bare string to avoid ruby >= 1.7 warning
    @artist, @title = response['DTITLE'].split(/ \/ /, 2)
    # A self-titled album may not have a title part
    @title ||= @artist

    response.each { |key, val|
      case key
	when /^4\d\d$/
	  raise(FreedbError, val)
	when /^TTITLE(\d+)$/
	  i = $1.to_i
	  @tracks_ext[i]["title"] = @tracks[i] = val
	when /^EXTT(\d+)$/
	  i = $1.to_i
	  @tracks_ext[i]["ext"] = val
      end
    }

    self
  end
  
  def close
    @handler.close if @handler
  end

private

  def _fetch
    @handler.gets #banner
    @handler.send_cmd("hello", "#{ENV['USER']} #{`hostname`.strip} ruby-freedb #{VERSION}")
    if @handler.gets.strip =~ /^4\d\d (.+)/ #welcome
      raise(FreedbError, $1)
    end
    set_proto_level(PROTO_LEVEL)
    @handler.send_cmd("query", @query)
    resp = @handler.gets.strip
    @results = []
    case resp
      when /^200 (.+)/ #single result
        @results << $1
      when /^211/, /^210/
        @handler.each_line { |l|
          @results << l
        }
      when /^202/
       #raise(FreedbError, "No match found")
       puts "no match found"
    end
    self
  end

  def set_proto_level(l)
    if l < 1
      raise(FreedbError, "Server doesn't support level 1!")
    end
    @handler.send_cmd("proto", l.to_s)
    if @handler.gets =~ /^501/
      set_proto_level(l-1)
    end
  end
  
  class FetchBase

    def each_line
      until (l = gets) =~ /^\./
        yield l.strip
      end
    end

  end

  class FetchCGI < FetchBase

    def initialize(server, port, proxy, proxy_port, url)
      require "net/http"
      require "cgi"
      @session = Net::HTTP.new(server, port, proxy, proxy_port)
      @url = url
      @proto_level = 1
    end

    def send_cmd(cmd, args)
      if cmd == "hello" 
        @hello_str = "hello=" + CGI.escape(args)
      elsif cmd == "proto"
        @proto_level = args
	@res = ["200"] #necessary for the next call to gets
      else
        request = "?cmd=cddb+#{cmd}+#{CGI.escape(args)}&#{@hello_str}&proto=#{@proto_level}"
        resp, data = @session.get(@url + request)
	raise(FreedbError, "Bad HTTP response: #{resp.code} #{resp.message} while querying server") unless resp.code == "200"
	@res = data.split("\n")
      end
    end

    def gets
      (@res.nil? ? "" : @res.shift)
    end
 
    def close
      #@session.finish
    end
  end

  class FetchNet < FetchBase

    def initialize(server, port)
      require "socket"
      @socket = TCPSocket.new(server, port)
    end

    def send_cmd(cmd, args)
      @socket.puts("cddb #{cmd} #{args}")
    end
    
    def gets
      @socket.gets
    end

    def close
      @socket.close      
    end 
  end

=begin
  class FetchDisk
    def initialize(dir)
      require "find"
      raise(FreedbError, "#{dir} is not a directory") if not dir.stat.directory?
      @dir = dir
    end

    def send_cmd(cmd, args)
      case cmd
        when "hello"
        when "query"
	  match = []
	  Find.find(@dir) { |f|
	    if md = /#{f}\/(\s+)\/#{args}/.match(f)
	      match << md[1] 
	    end
	  }
	  if match.size == 1
            @res[0] = "200"
	  elsif match.size > 1
	    @res[0] = "202"
	  else
	    @res[0] = "211"
	  end
        when "read"
	  categ, id = cmd.split(" ")
	  File.open(@dir + File::SEPARATOR + categ + File::SEPARATOR + id, "r") { |f|
	    @res = f.readlines
	  }
      end
    end

    def gets
      return @res.shift
    end

    def close
    end
  end
=end
end
